"""
Non-interferometric Lens Testing.

Documentation and examples are at <https://lenstest.readthedocs.io>

"""
__version__ = '0.9.0'
__author__ = 'Scott Prahl'
__email__ = 'scott.prahl@oit.edu'
__copyright__ = '2021-23, Scott Prahl'
__license__ = 'MIT'
__url__ = 'https://github.com/scottprahl/lenstest'

from .lenstest import *
from . import ronchi
from . import foucault
